# frozen_string_literal: true

require_relative 'utils'

class Danger::DangerWCC < Danger::Plugin
  class YarnDeduplicate
    include Utils

    DEFAULT_OPTIONS = {
      lockfile: 'yarn.lock',
      severity: :warn
    }.freeze

    def yarn_info
      @yarn_info ||=
        Danger::DangerWCC::Util::YarnInfo.new(self, @options)
    end

    def initialize(plugin, options = {})
      @plugin = plugin
      @options = DEFAULT_OPTIONS.merge(options)
    end

    def perform
      diff = run_yarn_deduplicate_diff

      each_addition_in_diff(diff) do |line|
        add_deduplicate_warning(line)
      end
    end

    private

    def run_yarn_deduplicate_diff
      diff =
        run_and_diff do
          run('npx yarn-deduplicate -s fewer --list '\
            "#{@options[:lockfile]}")
        end
      GitDiff.from_string(diff)
    end

    # rubocop:disable Layout/LineLength
    DEDUPE_REGEXP = /^\+Package \"([^\"]+)\" wants (\S+) and could get (\S+), but got (\S+)$/i.freeze
    # rubocop:enable Layout/LineLength

    def add_deduplicate_warning(line) # rubocop:disable Metrics/MethodLength
      warning = line.content.match(DEDUPE_REGEXP)
      return unless warning

      pkg, _semver, wanted, got = warning.captures

      idx = yarn_info.find_index_in_lockfile(
        pkg, got
      )
      msg = <<~HEREDOC
        You have an opportunity to deduplicate "#{pkg}".
        It's using #{got} but could use the existing version #{wanted}.

        You can fix this by running :
        `npx yarn-deduplicate -s fewer --packages "#{pkg}"`
      HEREDOC
      plugin.public_send(@options[:severity],
                         msg,
                         file: @options[:lockfile],
                         line: idx)
    end
  end
end

require_relative './util/yarn_info'
